#include <Wire.h>
#include <Servo.h>
#include <LiquidCrystal_I2C.h>
#include <Keypad.h>
#include <Adafruit_I2CDevice.h>

// Pinii unde sunt conectate componentele pe placuta Arduino
#define SERVO_MONEY_PIN 6
#define SERVO_RESET_PIN 7
#define BUTTON_MONEY_PIN 4 // Conectat la pinul D4
#define BUTTON_RESET_PIN 5 // Conectat la pinul D5
#define BUZZER_PIN 3
#define S0 8 // Pin pentru scalarea frecventei S0
#define S1 9 // Pin pentru scalarea frecventei S1
#define S2 10 // Pin pentru selectarea fotodiodelor S2
#define S3 11 // Pin pentru selectarea fotodiodelor S3
#define SENSOR_OUT 12 // Pin pentru iesirea semnalului de la senzor
#define RED_LED_PIN A3 // Pin pentru LED-ul roșu (alarmă)
#define GREEN_LED_PIN 3 // Pin pentru LED-ul verde

// Pozitiile de start ale servomotoarelor
#define SERVO_MONEY_START_POS 125
#define SERVO_RESET_START_POS 100

// Definire componente
Servo servo_money;
Servo servo_reset;
LiquidCrystal_I2C lcd(0x3F, 16, 2); // Adresă I2C pentru LCD

// Definire matrice tastatură și pini
const int ROW_NUM = 3; // Trei randuri
const int COLUMN_NUM = 3; // Trei coloane

char keys[ROW_NUM][COLUMN_NUM] = {
  {'1', '2', '3'},
  {'4', '5', '6'},
  {'7', '8', '9'}
};

byte pin_rows[ROW_NUM] = {A0, A1, A2}; // Randurile tastaturii
byte pin_column[COLUMN_NUM] = {A3, 2, 13}; // Coloanele tastaturii

Keypad keypad = Keypad(makeKeymap(keys), pin_rows, pin_column, ROW_NUM, COLUMN_NUM);

const String reset_password = "1234"; // Parola pentru reset
String input_password = "";

// Variabilele folosite in program
int pos_money = SERVO_MONEY_START_POS;
int pos_reset = SERVO_RESET_START_POS;
volatile int activate_sensor = 0;
volatile int reset = 0;
int move_money = 0;
int move_reset = 0;
int direction_money = 0;
int direction_reset = 0;
int red_freq = 0;
int green_freq = 0;
int blue_freq = 0;
int money_value = 0;
int sum = 0;
int attempts = 0;
bool alarm_active = false;
bool entering_pin = false;
bool system_locked = false;

// Declararea funcțiilor înainte de utilizare
void scanBill();
void resetSystem();
void getColors();
int check_money_type();
void playTone(int frequency, int duration);
void activateAlarm();
void resetSystemState();

void setup() {
  // Setup pentru servomotoare
  servo_money.attach(SERVO_MONEY_PIN);
  servo_money.write(0);
  servo_reset.attach(SERVO_RESET_PIN);
  servo_reset.write(0);

  // Setarea servomotoarelor in pozitiile initiale
  delay(150);
  servo_money.write(pos_money);
  servo_reset.write(pos_reset);

  // Setup pentru senzorul de culoare
  pinMode(S0, OUTPUT);
  pinMode(S1, OUTPUT);
  pinMode(S2, OUTPUT);
  pinMode(S3, OUTPUT);
  pinMode(SENSOR_OUT, INPUT);

  digitalWrite(S0, HIGH);
  digitalWrite(S1, LOW);

  // Setup pentru LCD
  lcd.init();
  lcd.backlight();
  lcd.clear();
  lcd.print("Total: ");
  lcd.print(sum);
  lcd.setCursor(0, 1);
  lcd.print("Insert Bill");

  // Setup pentru buzzer
  pinMode(BUZZER_PIN, OUTPUT);

  // Setup pentru LED-uri
  pinMode(RED_LED_PIN, OUTPUT);
  pinMode(GREEN_LED_PIN, OUTPUT);

  // Setup pentru butoane
  pinMode(BUTTON_MONEY_PIN, INPUT_PULLUP);
  pinMode(BUTTON_RESET_PIN, INPUT_PULLUP);

  // Initializare comunicatie seriala
  Serial.begin(9600);
}

void loop() {
  // Verificare dacă alarma este activă
  if (alarm_active) {
    digitalWrite(RED_LED_PIN, HIGH);
    digitalWrite(BUZZER_PIN, HIGH); // Activează buzzer-ul continuu
    lcd.clear();
    lcd.print("Alarm!");
    delay(10000); // Așteaptă 10 secunde
    digitalWrite(BUZZER_PIN, LOW);
    digitalWrite(RED_LED_PIN, LOW);
    alarm_active = false;
    entering_pin = false;
    system_locked = false;
    lcd.clear();
    lcd.print("Total: ");
    lcd.print(sum);
    lcd.setCursor(0, 1);
    lcd.print("Insert Bill");
    return;
  }

  // Verificare dacă se introduce PIN-ul
  if (entering_pin) {
    char key = keypad.getKey();
    if (key) {
      Serial.println(key);
      input_password += key; // Adaugă caracterul introdus la parola curentă
      lcd.clear();
      lcd.print("Enter PIN: ");
      lcd.print(input_password);
    }

    // Verificare dacă PIN-ul introdus are lungimea de 4 caractere
    if (input_password.length() == 4) {
      if (reset_password == input_password) {
        Serial.println("Reset password correct");
        sum = 0;
        lcd.clear();
        lcd.print("Total: ");
        lcd.print(sum);
        lcd.setCursor(0, 1);
        lcd.print("Reset");

        // Începe mișcarea servomotorului care oferă acces la zona de depozitare
        move_reset = 1;
        input_password = ""; // Șterge parola introdusă
        attempts = 0; // Resetează numărul de încercări
        entering_pin = false;
        system_locked = false; // Deblochează sistemul
      } else {
        Serial.println("Password incorrect for reset");
        attempts++;
        input_password = ""; // Resetează parola introdusă

        if (attempts >= 3) {
          activateAlarm();
        } else {
          lcd.clear();
          lcd.print("Wrong PIN!");
          lcd.setCursor(0, 1 );
          lcd.print("Try again!");
          delay(2000);
          lcd.clear();
          lcd.print("Enter PIN: ");
        }
      }
    }
  } else {
    // Verificare butoane
    int moneyButtonState = digitalRead(BUTTON_MONEY_PIN);
    int resetButtonState = digitalRead(BUTTON_RESET_PIN);

    // Verificare dacă butonul de introducere bancnote este apăsat
    if (!system_locked && moneyButtonState == LOW) {
      scanBill();
    }

    // Verificare dacă butonul de resetare este apăsat
    if (resetButtonState == LOW) {
      lcd.clear();
      lcd.print("Enter PIN: ");
      entering_pin = true;
      system_locked = true; // Blochează sistemul în timpul introducerii PIN-ului
    }

    // Actualizarea servomotoarelor
    servo_money.write(pos_money);
    servo_reset.write(pos_reset);
    delay(50);

    // Verificare dacă senzorul este activat
    if (activate_sensor == 1) {
      // Afisare mesaje pe LCD
      lcd.clear();
      lcd.print("Total: ");
      lcd.print(sum);
      lcd.setCursor(0, 1);
      lcd.print("Bill Inserted");

      // Scanarea bancnotei
      getColors();

      // Debugging frecvențe RGB
      Serial.print("Red frequency: ");
      Serial.println(red_freq);
      Serial.print("Green frequency: ");
      Serial.println(green_freq);
      Serial.print("Blue frequency: ");
      Serial.println(blue_freq);

      // Determinare tip bancnotă
      money_value = check_money_type();

      // Debugging valoarea bancnotei
      Serial.print("Detected money value: ");
      Serial.println(money_value);

      delay(1000);

      // Actualizarea sumei
      sum += money_value;

      // Debugging suma actualizată
      Serial.print("Updated sum: ");
      Serial.println(sum);

      lcd.clear();
      lcd.print("Total: ");
      lcd.print(sum);
      lcd.setCursor(0, 1);
      lcd.print("Please Wait");

      // Începe mișcarea servomotorului care muta bancnota în zona de depozitare
      move_money = 1;
      activate_sensor = 0;
    }

    // Mișcarea servomotorului pentru bancnote
    if (move_money == 1) {
      // Miscarea servomotorului
      if ((pos_money <=  SERVO_MONEY_START_POS) && (direction_money == 0)) {
        pos_money -= 10;
      }
      if ((pos_money <=  SERVO_MONEY_START_POS) && (direction_money == 1)) {
        pos_money += 10;
      }

      if (pos_money ==  SERVO_MONEY_START_POS) {
        // Servomotorul a ajuns în poziția inițială și bancnota în zona de depozitare
        move_money = 0;
        playTone(500, 1000); // Activează buzzerul

        // Afișează mesajele corespunzătoare pe LCD
        lcd.clear();
        lcd.print("Total: ");
        lcd.print(sum);
        lcd.setCursor(0, 1);
        lcd.print("Insert Bill");

        direction_money = 0;
      }
      if (pos_money <= 5) {
        // Schimbă direcția de mișcare după ce a ajuns în poziția finală
        direction_money = 1;
        delay(100);
      }
    }

    // Resetarea sistemului
    if (reset == 1) {    
      // Resetează suma și afișează mesajele corespunzătoare pe LCD
      sum = 0;
      lcd.clear();
      lcd.print("Total: ");
      lcd.print(sum);  
      lcd.setCursor(0, 1);
      lcd.print("Reset");

      // Începe mișcarea servomotorului care oferă acces la zona de depozitare
      move_reset = 1;
      reset = 0;
    }

    // Mișcarea servomotorului pentru resetare
    if (move_reset == 1) {
      // Mișcarea servomotorului
      if ((pos_reset >=  SERVO_RESET_START_POS) && (direction_reset == 0)) {
        pos_reset += 10;
      }
      if ((pos_reset >=  SERVO_RESET_START_POS) && (direction_reset == 1)) {
        pos_reset -= 10;
      }

      if (pos_reset ==  SERVO_RESET_START_POS) {
        // Servomotorul a ajuns în poziția inițială și bancnotele nu mai sunt în zona de depozitare
        move_reset = 0;
        playTone(500, 1000); // Activează buzzerul

        // Afișează mesajele corespunzătoare pe LCD
        lcd.clear();
        lcd.print("Total: ");
        lcd.print(sum);
        lcd.setCursor(0, 1);
        lcd.print("Insert Bill");

        direction_reset = 0;
      }
      if (pos_reset >= 175) {
        // Schimbă direcția de mișcare după ce a ajuns în poziția finală
        direction_reset = 1;
        delay(15000); // Așteaptă 15 sec
      }
    }
  }
}

void getColors() {
  digitalWrite(S2, LOW); // S2/S3 levels define which set of photodiodes we are using LOW/LOW is for RED, LOW/HIGH is for Blue, and HIGH/HIGH is for Green
  digitalWrite(S3, LOW);
  red_freq = pulseIn(SENSOR_OUT, LOW);
  red_freq = map(red_freq, 55, 353, 255, 0); // Remap
  delay(100);
  
  digitalWrite(S2, HIGH);
  digitalWrite(S3, HIGH); 
  green_freq = pulseIn(SENSOR_OUT, LOW);
  green_freq = map(green_freq, 55, 363, 255, 0); // Remap
  delay(100);
  
  digitalWrite(S2, LOW);
  digitalWrite(S3, HIGH);
  blue_freq = pulseIn(SENSOR_OUT, LOW);
  blue_freq = map(blue_freq, 46, 292, 255, 0); // Remap
  delay(100);

  Serial.print("R= ");
  Serial.print(red_freq);
  Serial.print(" G= ");
  Serial.print(green_freq);
  Serial.print(" B= ");
  Serial.print(blue_freq);
  Serial.println();
}

int check_money_type(){
  // Determinăm tipul de bancnotă pe baza frecvenței culorii predominante
  if (red_freq >= 250 || green_freq >= 250|| blue_freq >= 250 ) {
    return 0; // 50 lei
  }
  else if ((red_freq >= 160 && red_freq <= 191) && (green_freq >= 170 && green_freq <= 195) && (blue_freq >= 160 && blue_freq <= 192)) {
    return 100; // 100 lei
  }
  else if ((red_freq >= 110 && red_freq <= 220) && (green_freq >= 120 && green_freq <= 199) && (blue_freq >= 74 && blue_freq <= 190)) {
    return 50;
  }
 else if ((green_freq >= 238 && green_freq<= 250 ) || ( blue_freq >= 235 && blue_freq<=250)) {
    return 1; // 1 leu (actualizat)
  }
  else if ((red_freq >= 199 && red_freq <= 236) && (green_freq >= 199 && green_freq <= 227) && (blue_freq >= 190 && blue_freq <= 223)) {
    return 10; // 10 lei (actualizat)
  }
  else if ((red_freq >= 229) && (red_freq <= 249) && (blue_freq >= 223 && blue_freq <= 249)) {
    return 5; // 5 lei
  }
  return 0; // Bancnotă necunoscută
}

// Funcție pentru redarea unui ton la buzzer
void playTone(int frequency, int duration) {
  tone(BUZZER_PIN, frequency, duration);
  delay(duration);
  noTone(BUZZER_PIN);
}

// Funcție pentru activarea alarmei
void activateAlarm() {
  alarm_active = true;
  digitalWrite(RED_LED_PIN, HIGH);
  lcd.clear();
  lcd.print("Alarm!");
  playTone(1000, 10000); // Buzzer timp de 10 secunde
  digitalWrite(RED_LED_PIN, LOW);
  attempts = 0;
  lcd.clear();
  lcd.print("Total: ");
  lcd.print(sum);
  lcd.setCursor(0, 1);
  lcd.print("Insert Bill");
  alarm_active = false;
  system_locked = false;
  entering_pin = false;
}

// Funcție pentru scanarea unei bancnote
void scanBill() {
  activate_sensor = 1;
}

// Funcție pentru resetarea sistemului
void resetSystem() {
  reset = 1;
}
